! A small database/hash search module for use in renumbering programs
!
!Ignacy Misztal, University of Georgia
! 06/04/2001 - 02/07/2005
module hash_db
use searchsubs
implicit none


interface insert_el
   module procedure insert_el_int,insert_el_char,insert_el_real
end interface   

interface get_el
   module procedure get_el_int,get_el_char,get_el_real
end interface   


interface find_index
   module procedure find_index_char,find_index_int,find_index_real
end interface   


integer,private,parameter::maxfields=30         !max number of fields

! database main structure
type db_hash
  integer::nel,& 	! number of elements of hash table
           nfill,&	! number of these elemens filled
           n_int,&	! number of integer elements in one row
           status	! status is one of following
           		!    0 - uninitialized
           		!    1 - accepting field definitions
           		!    2 - ready to accept entries
           		!    3 - sorted
           		
  integer::nfields			 ! number of fields         		
  character (8)::fieldname(maxfields)	 !names of fields
  
  ! description of fields. Columns hold the following entries:
  !  1- description of fields (1- alpha, 2 - integer, 3- single real
  !  2- length of field if alphanumeric
  !  3,4 - First and last positions of fields in  hash table
  integer::fieldpar(4,maxfields)	 
  
  integer::n_ind,&	! position of index variable
           current      ! address of current record
  
  integer,pointer::hashtable(:,:)
 end type

! deafult formats
 character (10),private::def_fmt_real='(g0.4,1x)', &
                         def_fmt_int='(i0,1x)'
			 
! default sie of the table			 			 
 integer::default_db_size=100	 
 

CONTAINS


subroutine init(x)
! intializes x
type (db_hash)::x

x%status=0
x%nfields=0
x%n_ind=0
x%fieldname=''
x%fieldpar=0
x%nel=0
x%nfill=0
x%n_int=0
x%current=0
nullify(x%hashtable)
end subroutine


subroutine reset(x)
! releases memory of a database
type (db_hash)::x

if (x%status >=2) deallocate(x%hashtable)
call init(x)
end subroutine



subroutine zero(x,n)
type (db_hash)::x
integer::n

select case(x%status)
   case (0)
      print*," db_hash: fields not defined"
      stop
   case (1)
      if (x%n_ind <=0) then
         print*,"db_hash: index not defined"
         stop
      endif   
      call align_db(x)
      x%nel=n
      allocate(x%hashtable(x%n_int,x%nel))
      x%hashtable=0
    case(2,3)
      x%nfill=0
      if (x%nel /= n) then
         deallocate(x%hashtable)
	 x%nel=n
	 allocate(x%hashtable(x%n_int,x%nel))
      endif         
      x%hashtable=0
    case default
      print*,"dbase: status>3 should not happen"
      stop
end select
end subroutine



subroutine add_field(x, name,type,n)
! defines field with given name of type "char" with length n, "int" or "real"
! 
type (db_hash)::x
character (*)::name,type
integer,optional::n         

if (x%status == 0) x%status=1

if (x%status >1) then
   print*,'add_fields: all fields already declared; name=',name
   stop
endif   

if (x%nfields >= maxfields) then
   print*,"dbhash:maximum number of fields:",maxfields
   stop
 endif
 
 x%nfields=x%nfields+1
 x%fieldname(x%nfields)=name
 if (type(1:3)== "cha") then
    x%fieldpar(1,x%nfields)=1
    if (present(n)) then
          x%fieldpar(2,x%nfields)=n
        else
           print*,"db_hash: length of field missing for field ",name
           stop
    endif
  elseif (type(1:3)== "int") then
     x%fieldpar(1,x%nfields)=2
     x%fieldpar(2,x%nfields)=1
  elseif (type(1:3)== "rea") then
     x%fieldpar(1,x%nfields)=3
     x%fieldpar(2,x%nfields)=1
  else
     print*,"db_hash: unknown field type: ",type
     stop   
 endif
 
 end subroutine
 
   

subroutine set_index(x,name)
! sets field name as the index
type (db_hash)::x
character (*)::name
integer::i,m=0

m=find_field(x,name)
if (m == 0) then
   print*,'db_hash: field ',name,' does not exist'
   stop
endif

if (x%status >1) then
   print*,'set_index: index already set'
   stop
endif   

if (x%n_ind ==0) then
      x%n_ind=m
    else
      print*,'db_hash: index has already been set to: ',x%fieldname(x%n_ind)
      stop
endif
end subroutine       



function find_field(x,name) result(m)
! finds position of field with a given name
type (db_hash)::x
character (*)::name
integer::i,m

m=0

do i=1,x%nfields
   if (x%fieldname(i) == name) then
      m=i
     exit
   endif  
enddo
end function  




subroutine assert_status(x,n,text)
! if n>=0 and status/= n, prints 'incorrect state in' text and stops
! if n<0 and status < abs(n), prints 'too small state in' text and stops
!
type (db_hash)::x
integer::n
character (*)::text

if (n > 0 .and. x%status /= n) then
   print*,'incorrect state for subroutine/function ',text
   print*,'the status is ',x%status,' and should be ',n
   stop
endif

if (n < 0 .and. x%status < -n) then
   print*,'incorrect state for subroutine/function ',text
   print*,'the status is ',x%status,' and should be >',n
   stop
endif


end subroutine   




subroutine align_db(x)
! translates field descriptions into physical order
type (db_hash)::x
integer::i,m

call assert_status(x,1,'align_db') 

! process index first; index field is always physically first 
x%fieldpar(3,x%n_ind)=1
if (x%fieldpar(1,x%n_ind) == 1) then
      m=int((3+x%fieldpar(2,x%n_ind))/4)
   else
      m=1
endif      
x%fieldpar(4,x%n_ind)=m
x%n_int=m

! all other fields

do i=1,x%nfields
   if (i /= x%n_ind) then
       x%fieldpar(3,i)=x%n_int+1
       if (x%fieldpar(1,i) == 1) then
             m=int((3+x%fieldpar(2,i))/4)
          else
             m=1
      endif      
      x%fieldpar(4,i)=x%n_int+m
      x%n_int=x%n_int+m 
   endif
enddo
x%status=2
end subroutine



subroutine print_setup(x)
! print setup of the database
type (db_hash)::x
character (4)::c(3)=(/"char","int ","real"/)
integer:: i

select case (x%status)
   case (0)
      print*,'db_hash:  uninitialized'
      return
   case (1)
      print*,'db_hash: status:accepting field definitions'
   case (2)
      print*,'db_hash: status:accepting entries'
   case (3)
      print*,'db_hash: status:sorted'
end select     
   
print '(" field name  type   length   pos1  pos2")'
do i=1,x%nfields
   write(*,'(a9)',advance='no')x%fieldname(i)
   write(*,'(a5)',advance='no')c(x%fieldpar(1,i))
   write(*,'(3i8)')x%fieldpar(2:4,i)
enddo         
                
if (x%n_ind==0) then
      print*,'index not set'         
   else
      print*,' index variable: ',x%fieldname(x%n_ind)
endif

print*,'number of entries:',x%nel
print*,'number of filled entries:',x%nfill

end subroutine



subroutine write_setup(x,name)
! writes setup of the database to file 'name'

type (db_hash)::x
character (*)::name
integer::i

if (x%status <2 ) then
   print*,'write_setup: status < 2'
   stop
endif 

open(10,file=name)

write(10,'(8i10)')x%nfields,x%nel,x%n_int,x%nfields,x%n_ind
write(10,'(a)')x%fieldname(1:x%nfields)
write(10,'(4i10)')(x%fieldpar(:,i),i=1,x%nfields)
close(10)
end subroutine


subroutine write_data(x,name,form)
! writes content of the database to file 'name'
! form is "character" or "binary"; character by default

 type (db_hash)::x
character (*)::name
character,optional::form
logical::is_bin
integer::i

is_bin=.false.
if (present(form)) then
    if (form(1:1)=='b') then
       is_bin=.true.
    endif
endif       

if (x%status <2 ) then
   print*,'write_setup: status < 2'
   stop
endif 

if (is_bin) then
   open(10,file=name,form='unformatted')
  else
   open(10,file=name)
endif   

if (is_bin) then
      write(10)x%nfill
   else
      write(10,'(8i10)')x%nfill
endif
	
do i=1,x%nel
   if (x%hashtable(1,i) /= 0) then
       if (is_bin) then
             write(10)x%hashtable(:,i)  
	  else
	     write(10,'(6i12)')x%hashtable(:,i)
	endif
    endif
enddo
close(10)
end subroutine    	
	
!&&&&&&&

subroutine read_setup(x,name)
! reads setup of a database from file 'name'

type (db_hash)::x
character (*)::name
integer::i

if (x%status < 2 ) then
   call init(x)
  else
    deallocate(x%hashtable)
endif
x%status=1

open(10,file=name)


read(10,'(8i10)')x%nfields,x%nel,x%n_int,x%nfields,x%n_ind
read(10,'(a)')x%fieldname(1:x%nfields)
read(10,'(4i10)')(x%fieldpar(:,i),i=1,x%nfields)
x%status=1
close(10)
end subroutine


subroutine read_data(x,name,form,nel)
! reads content of the database from file 'name'
! form is "character" or "binary"; binary by default
! if optional 'nel' exists, the table will have 'nel elements;
!  otherwise will the table created is 1.25 larger than the number of elements

 type (db_hash)::x
character (*)::name
character,optional::form
logical::is_bin
integer::i,nind,nzero,io
integer,optional::nel
integer,allocatable::y(:)

is_bin=.false.
if (present(form)) then
    if (form(1:1)=='b') then
       is_bin=.true.
    endif
endif       

if (x%status /=1 ) then
   print*,'read_setup: setup not read'
   stop
endif 
x%status=2

if (is_bin) then
   open(10,file=name,form='unformatted')
  else
   open(10,file=name)
endif   

if (is_bin) then
      read(10)x%nfill
   else
      read(10,*)x%nfill
endif

if (present(nel)) then
    x%nel=nel
  else 
    x%nel=x%nfill/.8+2
endif
nzero=x%nfill; x%nfill=0
    
allocate(x%hashtable(x%n_int,x%nel))
x%hashtable=0

nind=x%fieldpar(4,1); allocate(y(x%n_int))
	
do i=1,nzero
    if (is_bin) then
          read(10,iostat=io)y 
        else
          read(10,*,iostat=io)y
    endif
    if (io /=0) then
       print*,' read_data unsuccessful'
       print*,'read ',i-1,' instead of ',nzero,' entries'
       stop
    endif
    call insert_index_int(x,y(1:nind))
    x%hashtable(nind+1:x%n_int,x%current)=y(nind+1:x%n_int)
enddo    

close(10)
deallocate(y)
end subroutine 

	     

  
subroutine insert_el_char(x,name,val)      
! inserts entry val to character field name into:
!   - new record if the field is an index
!   - current record otherwise   
type (db_hash)::x
character (*)::name,val
integer,allocatable::y(:)
integer::m,i1,i2

call assert_status(x,-2,'insert_el')

m=find_field(x,name)
if (m == 0) then
   print*,'db_hash: field ',name,' nonexistent'
   stop
endif

if (x%fieldpar(1,m) /= 1) then
   print*,'insert_el_char: field ',name,' not alphanumeric'
   stop
endif


i1=x%fieldpar(3,m); i2=x%fieldpar(4,m)
allocate(y(i2-i1+1))   !y keeps value in integer form
! pad string with spaces to prevnt garbage if string too short
y=transfer(trim(val)//repeat(' ',size(y)*4-len(trim(val))),y)

if (m /= x%n_ind) then
   ! insertion of non-index field into current record
   if (x%current == 0)then
      print*,'db_hash: insertion of non-index field before any index entry'
      stop
   endif  
   
   i1=x%fieldpar(3,m); i2=x%fieldpar(4,m)
   x%hashtable(i1:i2,x%current)=y

 else
   ! index; create new record
     call insert_index_int(x,y)
endif 

deallocate(y)
end subroutine


  
subroutine insert_el_int(x,name,val)      
! inserts entry val to integer field name into:
!   - new record if the field is an index
!   - current record otherwise   
type (db_hash)::x
character (*)::name
integer::val,m,i1,i2,y(1)

call assert_status(x,-2,'insert_el')

m=find_field(x,name)
if (m == 0) then
   print*,'db_hash: field ',name,' nonexistent'
   stop
endif

if (x%fieldpar(1,m) /= 2) then
   print*,'insert_el_char: field ',name,' not integer'
   stop
endif

!
if (m /= x%n_ind) then
   ! insertion of non-index field into current record
   if (x%current == 0)then
      print*,'db_hash: insertion of non-index field before any index entry'
      stop
   endif  
   
   
   x%hashtable(x%fieldpar(3,m),x%current)=val

 else
   ! index; create new record
     y=val
     call insert_index_int(x,y)
endif 
end subroutine



subroutine insert_el_real(x,name,val)      
! inserts entry val to real field name into:
!   - new record if the field is an index
!   - current record otherwise   
type (db_hash)::x
character (*)::name
integer::m,i1,i2,y
real::val

call assert_status(x,-2,'insert_el')
m=find_field(x,name)
if (m == 0) then
   print*,'db_hash: field ',name,' nonexistent'
   stop
endif

if (x%fieldpar(1,m) /= 3) then
   print*,'insert_el_char: field ',name,' not real'
   stop
endif

!
if (m /= x%n_ind) then
   ! insertion of non-index field into current record
   if (x%current == 0)then
      print*,'db_hash: insertion of non-index field before any index entry'
      stop
   endif  
   
   y=transfer(val,y)
   x%hashtable(x%fieldpar(3,m),x%current)=y

 else
   ! index; create new record
     y=transfer(val,y)
     call insert_index_int(x,(/y/))
endif 
end subroutine




subroutine insert_index_int(x,y)
! insert index y into database x

  interface	!external function
      function hashv1(dat,r,ind,m,n,mode,nr)
      implicit none
      integer ::hashv1,r,mode,nr,m,n
      integer::ind(:,:),dat(r)
      end function
  end interface
  
type (db_hash)::x
integer::y(:)   

call assert_status(x,2,'insert_el')

x%current=hashv1(y,size(y),x%hashtable,x%nel,x%n_int,1,x%nfill)
if (x%nfill > .7*x%nel) then
   call expand(x,2*x%nel)
   ! old address obsolete; find the current one
   x%current=hashv1(y,size(y),x%hashtable,x%nel,x%n_int,1,x%nfill)
endif  
end subroutine          

  
  
subroutine expand(x,n)
! expands the database twice, or if the optional variable n is present,
!  to size n
type (db_hash)::x
integer,optional::n
integer::nels

if (present(n)) then
     nels=n
   else
     nels=x%nel*2
endif     

call write_setup(x,'db_hash_expansion_setup')
call write_data(x,'db_hash_expansion_data')  
print*,'table expanded from ', x%nel,' to ',nels,' records'
call reset(x)
call read_setup(x,'db_hash_expansion_setup')
call read_data(x,'db_hash_expansion_data',nel=nels)  
  
open(10,file='db_hash_expansion_setup')  
close(10,status='delete')
open(10,file='db_hash_expansion_data')  
close(10,status='delete')  
x%status=2
end subroutine
  
  
  
subroutine get_el_char(x,name,val)  
! retrieves character field name from current record into val 
type (db_hash)::x
character (*)::name,val
integer::m,i1,i2,nc

call assert_status(x,2,'insert_el')

m=find_field(x,name)
if (m == 0) then
   print*,'db_hash: field ',name,' nonexistent'
   stop
endif

if (x%fieldpar(1,m) /= 1) then
   print*,'insert_el_char: field ',name,' not alphanumeric'
   stop
endif

if (x%current == 0) then
   print*,'get_el_char: current = 0'
   stop
endif

nc=x%fieldpar(2,m);i1=x%fieldpar(3,m); i2=x%fieldpar(4,m)

val=transfer(x%hashtable(i1:i2,x%current), val(1:nc))

end subroutine


  
subroutine get_el_int(x,name,val)      
! retrieves integer field name from current record into val
type (db_hash)::x
character (*)::name
integer::val,m,i1,i2,y(1)

call assert_status(x,2,'insert_el')

m=find_field(x,name)
if (m == 0) then
   print*,'get_el_int: field ',name,' nonexistent'
   stop
endif

if (x%fieldpar(1,m) /= 2) then
   print*,'get_el_char: field ',name,' not integer'
   stop
endif

if (x%current == 0) then
   print*,'get_el_integer: current = 0'
   stop
endif

val=x%hashtable(x%fieldpar(3,m),x%current)

end subroutine



subroutine get_el_real(x,name,val)      
! retrieves real field name from current record into val
type (db_hash)::x
character (*)::name
integer::m,i1,i2,y(1)
real::val

call assert_status(x,2,'insert_el')

m=find_field(x,name)
if (m == 0) then
   print*,'db_hash: field ',name,' nonexistent'
   stop
endif

if (x%fieldpar(1,m) /= 3) then
   print*,'insert_el_char: field ',name,' not real'
   stop
endif

if (x%current == 0) then
   print*,'get_el_real: current = 0'
   stop
endif

val=transfer(x%hashtable(x%fieldpar(3,m),x%current),1.0)

end subroutine




function find_index_int_vec(x,y) result (stat)
! find record with integer index y; y is a vector
! if successful return .true., otherwise .false.

  interface	!external function
      function hashv1(dat,r,ind,m,n,mode,nr)
      integer ::hashv1,r,mode,nr,m,n
      integer::ind(:,:),dat(r)
      end function
  end interface
   
type (db_hash)::x
integer::y(:)   
logical::stat

call assert_status(x,2,'insert_el')

x%current=hashv1(y,size(y),x%hashtable,x%nel,x%n_int,0,x%nfill)

stat=(x%current /= 0)

end function         



function find_index_int(x,y) result (stat)
! find record with integer index y; y is  a scalar
type (db_hash)::x
integer::y
logical::stat

stat=find_index_int_vec(x,(/y/))
end function



function find_index_char(x,y) result (stat)
! find record with character index y
type (db_hash)::x
character (*)::y
logical::stat
integer::i1,i2
integer,allocatable::z(:)

i1=x%fieldpar(3,1); i2=x%fieldpar(4,1)
allocate(z(i2-i1+1))	!z keeps value in integer form
! pad string with spaces to prevent garbage if string too short
z=transfer(trim(y)//repeat(' ',size(z)*4-len(trim(y))),z)

stat=find_index_int_vec(x,z)

deallocate(z)
end function



function find_index_real(x,y) result (stat)
! find record with real index z
type (db_hash)::x
real::y
logical::stat
integer::z(1)

z=transfer(y,z)

stat=find_index_int_vec(x,z)
end function




recursive function move_rec(x,direction) result (status)
! moves the current record according to alphanumerical variable direction:
!  "next", "previous","first","last"
! variable status is .true. if the operation is successful
!
type (db_hash)::x
character (*)::direction
logical::status
integer::i,i1,i2

if (x%status < 2) then
   print*,'incorrect state for move_rec'
   print*,'current state ',x%status
   stop
endif   

! entry in hash table is empty if index = 0

i1=x%fieldpar(3,1); i2=x%fieldpar(4,1)	!index delimiters

status=.false.

select case (direction(1:1))
   case("f")		!"first", which is first nonzero from the beginning
      x%current=0
      status=move_rec(x,'n')
   case ("l")		!"last" which is first nonzero below the top   
      x%current=x%nel+1
      status=move_rec(x,'p')
   case ("n")		!next
      do i=x%current+1,x%nel
         if (any(x%hashtable(i1:i2,i) /=0)) then
            x%current=i
            status=.true.
	    exit
         endif
      enddo
   case ("p")		!previous
      do i=x%current-1,1,-1
         if (any(x%hashtable(i1:i2,i) /=0)) then
            x%current=i
            status=.true.
	    exit
         endif
      enddo  
   case default
      print*,'move_rec: unknown direction:',direction
end select  
end function


subroutine print_entry(x,form,unit)
! prints the current entry either with given format (form) or with default
! format; if optional unit is present, printing is to that unit; otherwise
! printing is to a terminal
type (db_hash)::x
integer::i,nc,i1,i2,un1
character (50)::c
real::r
character (*),optional::form
integer,optional::unit

if (present(unit)) then
   un1=unit
  else
   un1=6
endif    

if (x%status <2) then
   print*,'print_entry: empty table'
endif   

 if (present(form)) then
   write(un1,form)x%hashtable(:,x%current)
  else 
   do i=1,x%nfields
      nc=x%fieldpar(2,i); i1=x%fieldpar(3,i); i2=x%fieldpar(4,i)
      !print*,nc,i1,i2,x%fieldpar(1,i)
      !print*,x%fieldpar(1,x%current)
      !print*,x%fieldname(i)
      select case (x%fieldpar(1,i))
         case(1)	!alphanumeric
           write(un1,'(a,1x)',advance='no')&
                      trim(transfer(x%hashtable(i1:i2,x%current),c(1:nc)))
         case(2)	!integer
           write(un1,def_fmt_int,advance='no')x%hashtable(i1,x%current)            
         case(3)	!real
           write(un1,def_fmt_real,advance='no')&
                      transfer(x%hashtable(i1,x%current),r)
      endselect
   enddo
   write (un1,*)
endif   
end subroutine


subroutine print_table(x,form,un,label)
! prints entire table with format if optional form
! present, or with default format
!
!output goes to unit un if present, or to a terminal
type (db_hash)::x
logical::stat
logical, optional:: label
character (*),optional::form
integer,optional::un
integer :: i


if (present(label) .and. label) then 
    do i=1,x%nfields
       if (present(un)) then
           write(un,'(a,x)',advance='no') trim(x%fieldname(i))
       else
           write(*,'(a,x)',advance='no') trim(x%fieldname(i))
       endif
    enddo
       if (present(un)) then
           write(un,'(a)') ''
       else
           write(*,'(a)') ''
       endif
    
endif
stat=move_rec(x,'first')
do
    if (.not. stat) exit
       if (present(un)) then
            call print_entry(x,form,un)
         else
           call print_entry(x,form)
       endif
    stat=move_rec(x,'next')
enddo

end subroutine                     


subroutine add_count(x,field,type)
! Adds consecutive numbering from 1 to given field;
! If optional type='sorted', then entries are numbered by sorted
! indexes. Otherwise, numbering is by random order.
! 
! If the option 'sorted is used', the table is sorted and cannot be used 
! for insertion. To rehash, use "call expand(x,x%nel)"

type (db_hash)::x
character (*)::field
character (*),optional::type
integer::cons,i
logical::stat,srt=.false.
                                                                                
   if (present(type)) then
        if (type(1:2)=='so') then
           call sort_db(x)
           print*,'table with ',x%nfill,' elements sorted'
           srt=.true.
        endif
   endif

cons=0
stat=move_rec(x,'first')
do
    if (.not. stat) exit
    cons=cons+1
    call insert_el(x,field,cons)
    stat=move_rec(x,'next')
enddo

end subroutine


subroutine sort_db(x,fields,order)
! sorts database x specified by optional fields and order
! if optional arguments are missing, it sorts by index in ascending order
type (db_hash)::x
character (*),optional::fields,order
character (50)::types='',vc(20)*50
integer::m,n,item,v(50),p(4),i,j,col(20)


if (x%status <2) then
   print*,'sort_db: database not set yet'
   stop
endif   

if (present(fields)) then 
      call tokenize(fields,vc,n) 
  else
      vc(1)=x%fieldname(1); n=1
endif    
 
      item=0
      do i=1,n  
         m=find_field(x,vc(i))
         if (m==0) then
            print*,'sort_db: field ',trim(vc(i)),' not found'
            stop
         endif
         p=x%fieldpar(:,m)
         select case (p(1))
           case(1)		!alphanumeric
             col(item+1:item+p(4)-p(3)+1)=(/(j,j=p(3),p(4))/)
	     types=trim(types)//repeat('a',p(4)-p(3)+1)
	     item=item+p(4)-p(3)+1
           case(2)		!integer
	     col(item+1)=p(3)
	     types=trim(types)//'i'
	     item=item+1
	   case(3)
	     col(item+1)=p(3)
	     types=trim(types)//'r'
	     item=item+1  
	   case default
	     print*,'This should not happen QF234TYr'
         end select
      enddo   	     
      
call sortmat(x%hashtable,col(1:item),order=order,types=types)

x%status=3

end subroutine

 
 


end module
